#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import copy
import functools
import re

from .errors import *

class Simulation():
	'''
	Represent a simulation, itself identified by its settings.

	Parameters
	----------
	folder : Folder
		Instance of `Folder` for the current simulation's folder.

	settings : dict
		Dictionary listing the user settings.
	'''

	def __init__(self, folder, settings):
		self._folder = folder
		self._user_settings = settings

		self._raw_globalsettings = None
		self._raw_settings = None
		self._raw_settings_dict = None

		self._setting_tag_regex_compiled = None
		self._parser_recursion_stack = []

	@classmethod
	def ensureType(cls, simulation, folder):
		'''
		Ensure a variable is a Simulation object.

		Parameters
		----------
		simulation : dict|Simulation
			The variable to check.

		folder : Folder
			The `Folder` instance to use in case we need to create a new object.

		Returns
		-------
		simulation : Simulation
			The simulation as a Simulation instance.
		'''

		if type(simulation) is cls:
			return simulation

		return cls(folder, simulation)

	def __getitem__(self, key):
		'''
		Access to a global setting.

		Parameters
		----------
		key : str
			The key of the setting to get.

		Raises
		------
		KeyError
			The key does not exist.

		Returns
		-------
		value : mixed
			The corresponding value.
		'''

		try:
			return self.reduced_globalsettings[key]

		except KeyError:
			raise KeyError('The key does not exist in the global settings')

	def __setitem__(self, key, value):
		'''
		Change a global setting.

		Parameters
		----------
		key : str
			The key of the setting to change.

		value : mixed
			The new value of the setting.

		Raises
		------
		KeyError
			The key does not exist.
		'''

		try:
			setting = [setting for setting in self._globalsettings if setting['name'] == key][0]

		except IndexError:
			raise KeyError('The key does not exist in the global settings')

		else:
			setting['value'] = value

	@property
	def _globalsettings(self):
		'''
		Return (and generate if needed) the complete list of global settings.

		Returns
		-------
		raw_globalsettings : list
			The global settings.
		'''

		if not(self._raw_globalsettings):
			self.generateGlobalSettings()

		return self._raw_globalsettings

	@property
	def _settings(self):
		'''
		Return (and generate if needed) the complete list of settings.

		Returns
		-------
		raw_settings : list
			The settings.
		'''

		if not(self._raw_settings):
			self.generateSettings()

		return self._raw_settings

	@property
	def _settings_dict(self):
		'''
		Return (and generate if needed) the complete list of settings as a dictionary.

		Returns
		-------
		raw_settings : dict
			The settings.
		'''

		if not(self._raw_settings_dict):
			self.generateSettings()

		return self._raw_settings_dict

	@property
	def settings(self):
		'''
		Return the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : list
			List of sets of settings.
		'''

		return [
			{s.name: s.value for s in settings_set if not(s.exclude)}
			for settings_set in self._settings
		]

	@property
	def settings_dict(self):
		'''
		Return a dictionary with the complete list of sets of settings to use, as dictionaries.
		The settings with `exclude` to `True` are ignored.

		Returns
		-------
		settings : dict
			List of sets of settings.
		'''

		return {
			settings_set_name: [
				{s.name: s.value for s in settings_set if not(s.exclude)}
				for settings_set in settings_sets
			]
			for settings_set_name, settings_sets in self._settings_dict.items()
		}

	@property
	def settings_as_strings(self):
		'''
		Return the complete list of sets of settings to use, as strings.
		Take into account the `only_if` parameter.

		Returns
		-------
		settings : list
			Settings, generated according to their pattern.
		'''

		return [
			[str(s) for s in settings_set if s.shouldBeDisplayed()]
			for settings_set in self._settings
		]

	@property
	def reduced_globalsettings(self):
		'''
		Return the list of global settings, as a name: value dictionary.

		Returns
		-------
		settings : dict
			The global settings.
		'''

		return {setting['name']: setting['value'] for setting in self._globalsettings}

	@property
	def reduced_settings(self):
		'''
		Return the list of settings, as a name: value dictionary.
		Ignore multiple occurrences of the same setting.

		Returns
		-------
		settings : dict
			The settings.
		'''

		return functools.reduce(lambda a, b: {**a, **b}, self.settings)

	@property
	def command_line(self):
		'''
		Return the command line to use to generate this simulation.

		Returns
		-------
		command_line : str
			The command line to execute.
		'''

		return ' '.join([self._folder.settings['exec']] + sum(self.settings_as_strings, []))

	@property
	def _setting_tag_regex(self):
		'''
		Regex to detect whether there is a setting or global setting tag in a string.

		Returns
		-------
		regex : re.Pattern
			The setting tag regex.
		'''

		if self._setting_tag_regex_compiled is None:
			self._setting_tag_regex_compiled = re.compile(r'\{(?P<category>(?:global)?setting):(?P<name>[^}]+)\}')

		return self._setting_tag_regex_compiled

	def generateGlobalSettings(self):
		'''
		Generate the full list of global settings.
		'''

		self._raw_globalsettings = []

		for setting in self._folder.settings['globalsettings']:
			self._raw_globalsettings.append({
				'name': setting['name'],
				'value': self._folder.applyFixers(self._user_settings[setting['name']]) if setting['name'] in self._user_settings else setting['default']
			})

		self.parseGlobalSettings()

	def generateSettings(self):
		'''
		Generate the full list of settings, taking into account the user settings and the default values in the folder.
		The "raw settings" are generated.
		Each set of settings is a list of all settings in this set.
		'''

		if type(self._user_settings['settings']) is list:
			user_settings = {
				setting_set_name: [s['settings'] for s in self._user_settings['settings'] if s['set'] == setting_set_name]
				for setting_set_name in set([s['set'] for s in self._user_settings['settings']])
			}

		else:
			user_settings = self._user_settings['settings']

		self._raw_settings_dict = {}
		default_pattern = self._folder.settings['setting_pattern']

		for settings_set in self._folder.settings['settings']:
			default_settings = [
				SimulationSetting(self._folder, settings_set['set'], s['name'])
				for s in settings_set['settings']
			]

			try:
				values_sets = user_settings[settings_set['set']]

			except KeyError:
				if settings_set['required']:
					self._raw_settings_dict[settings_set['set']] = [default_settings]

			else:
				if not(type(values_sets) is list):
					values_sets = [values_sets]

				self._raw_settings_dict[settings_set['set']] = []

				for values_set in values_sets:
					set_to_add = copy.deepcopy(default_settings)

					for setting in set_to_add:
						try:
							setting.value = values_set[setting.name]

						except KeyError:
							pass

					self._raw_settings_dict[settings_set['set']].append(set_to_add)

		self._raw_settings = sum(self._raw_settings_dict.values(), [])
		self.parseSettings()

	def parseString(self, s):
		'''
		Parse a string to take into account possible settings.
		The tag `{setting:name}` is replaced by the value of the simulation's setting named `name`.
		The tag `{globalsetting:name}` is replaced by the value of the global setting named `name`.

		Tags are replaced recursively.

		Parameters
		----------
		s : str
			The string to parse.

		Returns
		-------
		parsed : mixed
			The parsed string, or a copy of the setting if the whole string is just one tag.
		'''

		if not(type(s) is str):
			return s

		settings = {
			'setting': self.reduced_settings,
			'globalsetting': self.reduced_globalsettings
		}

		fullmatch = self._setting_tag_regex.fullmatch(s)

		if fullmatch:
			try:
				return copy.deepcopy(settings[fullmatch.group('category')][fullmatch.group('name')])

			except KeyError:
				return s

		parsed = ''
		k0 = 0

		for match in self._setting_tag_regex.finditer(s):
			parsed += s[k0:match.start()]

			try:
				parsed += str(settings[match.group('category')][match.group('name')])

			except KeyError:
				parsed += match.group(0)

			k0 = match.end()

		parsed += s[k0:]

		self._parser_recursion_stack.append(s)

		if not(parsed in self._parser_recursion_stack):
			return self.parseString(parsed)

		self._parser_recursion_stack.clear()

		return parsed

	def parseGlobalSettings(self):
		'''
		Parse the global settings to take into account possible other settings' values.
		'''

		for setting in self._globalsettings:
			setting['value'] = self.parseString(setting['value'])

	def parseSettings(self):
		'''
		Parse the settings to take into account possible other settings' values.
		'''

		for settings_set in self._settings:
			for setting in settings_set:
				setting.value = self.parseString(setting.value)

class SimulationSetting():
	'''
	Represent a simulation setting.

	Parameters
	----------
	folder : Folder
		The simulations folder this setting is part of.

	set_name : str
		Name of the set this setting is part of.

	setting_name : str
		Name of this setting.

	Raises
	------
	SettingsSetNotFoundError
		The settings set has not been found.

	SettingNotFoundError
		The setting has not been found in the set.
	'''

	def __init__(self, folder, set_name, setting_name):
		self._folder = folder

		try:
			settings_set_dict = [
				settings_set
				for settings_set in self._folder.settings['settings']
				if settings_set['set'] == set_name
			][0]

		except IndexError:
			raise SettingsSetNotFoundError(set_name)

		try:
			setting_dict = [
				setting
				for setting in settings_set_dict['settings']
				if setting['name'] == setting_name
			][0]

		except IndexError:
			raise SettingNotFoundError(set_name, setting_name)

		self._name = setting_dict['name']
		self._value = setting_dict['default']
		self._exclude_for_db = 'exclude' in setting_dict and setting_dict['exclude']
		self._pattern = setting_dict['pattern'] if 'pattern' in setting_dict else self._folder.settings['setting_pattern']

		self._use_only_if = 'only_if' in setting_dict
		if self._use_only_if:
			self._only_if_value = setting_dict['only_if']

		self._fixers = {'before': [], 'after': []}
		for when in ['before', 'after']:
			try:
				self._fixers[when] = setting_dict[f'fixes_{when}']

			except KeyError:
				pass

	def __str__(self):
		'''
		Return a string representing the setting and its value, according to the pattern.

		Returns
		-------
		setting : str
			The representation of the setting.
		'''

		return self._pattern.format(name = self.name, value = self.value)

	@property
	def name(self):
		'''
		Getter for the setting name.

		Returns
		-------
		name : str
			Name of the setting.
		'''

		return self._name

	@property
	def value(self):
		'''
		Getter for the setting value.

		Returns
		-------
		value : mixed
			Value of the setting.
		'''

		return self._value

	@value.setter
	def value(self, new_value):
		'''
		Setter for the setting value.

		Parameters
		----------
		new_value : mixed
			New value of the setting
		'''

		self._value = self._folder.applyFixers(new_value, **self._fixers)

	@property
	def exclude(self):
		'''
		Return `True` if the setting should be excluded to define a simulation.

		Returns
		-------
		exclude : bool
			Exclusion parameter.
		'''

		return self._exclude_for_db

	def shouldBeDisplayed(self):
		'''
		Return `True` if the setting can be displayed/used for the execution.
		This check is based on the `only_if` parameter.

		Returns
		-------
		should_be_displayed : bool
			The result of the check.
		'''

		return not(self._use_only_if) or self.value == self._only_if_value



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import datetime

from .errors import *

class UI():
	'''
	Represent the user interface. Make easy the display of lines that can be updated, and the creation of progress bars.
	'''

	def __init__(self, *, progress_bars_length = 40, progress_bars_empty_char = '░', progress_bars_full_char = '█'):
		self._cursor_vertical_pos = 0

		self._text_lines = {}

		self._progress_bars = {}
		self._progress_bars_length = progress_bars_length
		self._progress_bars_empty_char = progress_bars_empty_char
		self._progress_bars_full_char = progress_bars_full_char

	@property
	def _last_line(self):
		'''
		Get the position of the last line.

		Returns
		-------
		pos : int
			The position of the last line.
		'''

		return len(self._text_lines) + len(self._progress_bars)

	def moveCursorTo(self, pos):
		'''
		Move the cursor to a new vertical position.

		Parameters
		----------
		pos : int
			The position to go to.
		'''

		cursor_offset = pos - self._cursor_vertical_pos

		if cursor_offset != 0:
			cursor_direction = 'A' if cursor_offset < 0 else 'B'
			print(f'\u001b[{abs(cursor_offset)}{cursor_direction}', end = '\r')

			self._cursor_vertical_pos = pos

	def moveCursorRight(self, dx):
		'''
		Move the cursor to the right.

		Parameters
		----------
		dx : int
			The movement to execute.
		'''

		if dx > 0:
			print(f'\u001b[{dx}C', end = '')

	def addTextLine(self, text):
		'''
		Add a new text line to display.

		Parameters
		----------
		text : str
			The text to display

		Returns
		-------
		id : str
			The ID to use to refer to this new line.
		'''

		self.moveCursorTo(self._last_line)
		print(text)

		line_id = hex(int(datetime.datetime.now().timestamp() * 1E6))[2:]
		self._text_lines[line_id] = {
			'position': self._cursor_vertical_pos,
			'text': text,
			'length': len(text)
		}

		self._cursor_vertical_pos += 1

		return line_id

	def addProgressBar(self, N):
		'''
		Add a new progress bar.

		Parameters
		----------
		N : int
			The final number to reach to display the famous 100%.

		Returns
		-------
		id : str
			The ID to use to refer to this progress bar.
		'''

		self.moveCursorTo(self._last_line)

		pattern = f'{{n:>{len(str(N))}d}}/{{N:d}} {{bar:{self._progress_bars_empty_char}<{self._progress_bars_length}}} {{p:>6.1%}}'
		first_bar = pattern.format(n = 0, N = N, bar = '', p = 0)
		print(first_bar)

		bar_id = hex(int(datetime.datetime.now().timestamp() * 1E6))[2:]
		self._progress_bars[bar_id] = {
			'position': self._cursor_vertical_pos,
			'n': 0,
			'N': N,
			'pattern': pattern,
			'length': len(first_bar)
		}

		self._cursor_vertical_pos += 1

		return bar_id

	def replaceTextLine(self, id, new_text):
		'''
		Replace a text line.

		Parameters
		----------
		id : str
			The ID of the line to display.

		new_text : str
			The new text to display.

		Raises
		------
		UITextLineNotFoundError
			The ID does not refer to a known text line.
		'''

		try:
			text_line = self._text_lines[id]

		except KeyError:
			raise UITextLineNotFoundError(id)

		else:
			self.moveCursorTo(text_line['position'])

			print(' ' * text_line['length'], end = '\r')
			print(new_text, end = '\r')
			text_line['text'] = new_text
			text_line['length'] = len(new_text)

			self.moveCursorTo(self._last_line)

	def updateProgressBar(self, id, n = None):
		'''
		Update a progress bar.

		Parameters
		----------
		id : str
			The ID of the bar to update.

		n : int
			The new value of the counter. If `None`, increment the counter by one.

		Raises
		------
		UIProgressBarNotFoundError
			The ID does not refer to a known progress bar.
		'''

		try:
			progress_bar = self._progress_bars[id]

		except KeyError:
			raise UIProgressBarNotFoundError(id)

		else:
			self.moveCursorTo(progress_bar['position'])

			if n is None:
				n = progress_bar['n'] + 1

			percentage = n / progress_bar['N']
			length_N = len(str(progress_bar['N']))

			print(f'{{n:>{length_N}d}}'.format(n = n), end = '\r')

			dx = length_N * 2 + 2
			self.moveCursorRight(dx)
			print(self._progress_bars_full_char * round(percentage * self._progress_bars_length), end = '\r')

			self.moveCursorRight(dx + self._progress_bars_length + 1)
			print('{p:>6.1%}'.format(p = percentage), end = '\r')

			progress_bar['n'] = n

			self.moveCursorTo(self._last_line)

	def moveUp(self, line):
		'''
		Move a line to the line above, assuming the above line is empty.

		Parameters
		----------
		line : dict
			Line to move.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		if line['position'] <= 0:
			raise UINonMovableLine(line['position'])

		self.moveCursorTo(line['position'])
		print(' ' * line['length'], end = '\r')

		text = ''
		try:
			text = line['text']

		except KeyError:
			percentage = line['n'] / line['N']
			text = line['pattern'].format(n = line['n'], N = line['N'], bar = self._progress_bars_full_char * round(percentage * self._progress_bars_length), p = percentage)

		finally:
			self.moveCursorTo(line['position'] - 1)
			print(text, end = '\r')
			line['position'] -= 1

	def moveUpFrom(self, pos):
		'''
		Move all lines starting at a given position.

		Parameters
		----------
		pos : int
			Position to start from.

		Raises
		------
		UINonMovableLine
			The line can't be moved.
		'''

		if pos <= 0:
			raise UINonMovableLine(pos)

		lines_to_move = [line for line in [*self._text_lines.values(), *self._progress_bars.values()] if line['position'] >= pos]
		lines_to_move.sort(key = lambda line: line['position'])

		for line in lines_to_move:
			self.moveUp(line)

	def _removeLine(self, lines, id):
		'''
		Remove a line and move up all the lines below.

		Parameters
		----------
		lines : dict
			Dict where the lines are stored.

		id : str
			ID of the line to remove.

		Raises
		------
		UILineNotFoundError
			The ID does not refer to a known text line.
		'''

		try:
			line = lines[id]

		except KeyError:
			raise UILineNotFoundError(id)

		else:
			self.moveCursorTo(line['position'])
			print(' ' * line['length'], end = '\r')

			self.moveUpFrom(line['position'] + 1)

			del lines[id]
			self.moveCursorTo(self._last_line)

	def removeTextLine(self, line_id):
		'''
		Remove a text line and move all the lines below.

		Parameters
		----------
		line_id : str
			The ID of the line to remove.
		'''

		self._removeLine(self._text_lines, line_id)

	def removeProgressBar(self, line_id):
		'''
		Remove a progress bar and move all the lines below.

		Parameters
		----------
		line_id : str
			The ID of the line to remove.
		'''

		self._removeLine(self._progress_bars, line_id)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import shutil
import stat
import re
import functools
import inspect
import copy
from math import ceil
from string import Template

from .errors import *
from .simulation import Simulation
from . import generators

class Generator():
	'''
	Generate some scripts to create a set of simulations.
	Initialize the list of simulations to generate.

	Parameters
	----------
	folder : Folder
		The folder to manage. Must contain a settings file.
	'''

	def __init__(self, folder):
		self._folder = folder

		self._simulations_to_generate = []

		self._lists_regex_compiled = None
		self._lists_items_regex_compiled = None

		self._variables_generators_regex_compiled = None
		self._variables_generators_list = None

	@property
	def _variables_generators_regex(self):
		'''
		Regex to determine whether a function's name corresponds to a variable generator.

		Returns
		-------
		regex : re.Pattern
			The generators regex.
		'''

		if self._variables_generators_regex_compiled is None:
			self._variables_generators_regex_compiled = re.compile(r'^generator_([A-Za-z0-9_]+)$')

		return self._variables_generators_regex_compiled

	@property
	def _lists_regex(self):
		'''
		Regex to detect the presence of lists blocks in a skeleton.

		Returns
		-------
		regex : re.Pattern
			The lists regex.
		'''

		if self._lists_regex_compiled is None:
			self._lists_regex_compiled = re.compile(r'^[ \t]*#{3} BEGIN_(?P<tag>[A-Z_]+) #{3}$.+?^(?P<content>.+?)^[ \t]*#{3} END_\1 #{3}$.+?^', flags = re.MULTILINE | re.DOTALL)

		return self._lists_regex_compiled

	@property
	def _lists_items_regex(self):
		'''
		Regex to detect the presence of lists items as variables names.

		Returns
		-------
		regex : re.Pattern
			The lists items regex.
		'''

		if self._lists_items_regex_compiled is None:
			self._lists_items_regex_compiled = re.compile(r'^(?P<list>[A-Z_]+)__(?P<index>[0-9]+)$')

		return self._lists_items_regex_compiled

	@property
	def _variables_generators(self):
		'''
		Get the list of available variables generators.

		Returns
		-------
		generators : dict
			The generators.
		'''

		if self._variables_generators_list is None:
			self._variables_generators_list = {}
			self.loadVariablesGeneratorsFromModule(generators)

		return self._variables_generators_list

	def loadVariablesGeneratorsFromModule(self, module):
		'''
		Load all variables generators in a given module.

		Parameters
		----------
		module : Module
			Module (already loaded) where are defined the generators.
		'''

		for function in inspect.getmembers(module, inspect.isfunction):
			generator_match = self._variables_generators_regex.match(function[0])

			if generator_match:
				self.setVariableGenerator(generator_match.group(1), function[1])

	def setVariableGenerator(self, generator_name, generator):
		'''
		Set (add or replace) a variable generator.

		Parameters
		----------
		generator_name : str
			Name of the generator.

		generator : function
			Generator to register.
		'''

		self._variables_generators[generator_name] = generator

	def removeVariableGenerator(self, generator_name):
		'''
		Remove a variable generator.

		Parameters
		----------
		generator_name : str
			Name of the generator to remove.
		'''

		if not(generator_name in self._variables_generators):
			raise VariableGeneratorNotFoundError(generator_name)

		del self._variables_generators[generator_name]

	def add(self, simulations):
		'''
		Add simulations to generate.

		Parameters
		----------
		simulations : list|dict|Simulation
			List of simulations to add.
		'''

		if type(simulations) is list:
			for simulation in simulations:
				self._simulations_to_generate.append(Simulation.ensureType(simulation, self._folder))

		else:
			self.add([simulations])

	def clear(self):
		'''
		Clear the list of simulations to generate.
		'''

		self._simulations_to_generate.clear()

	def parse(self, simulations_set = None):
		'''
		Parse a set of simulations to generate the corresponding command lines and other variables.

		Parameters
		----------
		simulations_set : list
			The set of simulations to parse. If `None`, default to the whole list.

		Returns
		-------
		variables : dict
			The list of command lines, and variables corresponding to the simulations' global settings.
		'''

		if simulations_set is None:
			simulations_set = self._simulations_to_generate

		globalsettings = self._folder.settings['globalsettings']

		variables = {
			'data_lists': {'COMMAND_LINES': [simulation.command_line for simulation in simulations_set]},
			'data_variables': {}
		}

		for globalsetting in self._folder.settings['globalsettings']:
			name_upper = 'GLOBALSETTING_'+globalsetting['name'].upper()
			variables['data_lists'][name_upper] = [simulation[globalsetting['name']] for simulation in simulations_set]

			if 'generators' in globalsetting:
				for generator_name in globalsetting['generators']:
					if not(generator_name in self._variables_generators):
						raise VariableGeneratorNotFoundError(generator_name)

					variables['data_variables'][generator_name.upper()+'_'+name_upper] = functools.reduce(self._variables_generators[generator_name], variables['data_lists'][name_upper])

		return variables

	def generateScriptFromSkeleton(self, skeleton_name, output_name, lists, variables, *, make_executable = False):
		'''
		Generate a script from a skeleton, using a given set of command lines.

		Parameters
		----------
		skeleton_name : str
			Name of the skeleton file.

		output_name : str
			Name of the script to write.

		lists : dict
			Lists we can use to loop through.

		variables : dict
			Variables we can use in the whole script template.

		make_executable : boolean
			`True` to add the 'exec' permission to the script.
		'''

		with open(skeleton_name, 'r') as f:
			skeleton = f.read()

		script_content = ''
		k0 = 0

		for match in self._lists_regex.finditer(skeleton):
			if match.group('tag') in lists:
				script_content += skeleton[k0:match.start()]

				loop_content_template = Template(match.group('content'))
				for index, value in enumerate(lists[match.group('tag')]):
					script_content += loop_content_template.safe_substitute(ITEM_INDEX = index, ITEM_VALUE = value)

				k0 = match.end()

		script_content += skeleton[k0:]

		variables = copy.deepcopy(variables)
		script_content_template = Template(script_content)

		for match in script_content_template.pattern.finditer(script_content_template.template):
			if match.group('braced'):
				submatch = self._lists_items_regex.match(match.group('braced'))

				if submatch:
					variables[submatch.group(0)] = lists[submatch.group('list')][int(submatch.group('index'))]

		script_content = script_content_template.safe_substitute(**variables)

		with open(output_name, 'w') as f:
			f.write(script_content)

		if make_executable:
			os.chmod(output_name, os.stat(output_name).st_mode | stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH)

	def generate(self, dest_folder, recipe, *, empty_dest = False):
		'''
		Generate the scripts to launch the simulations by subgroups.

		Parameters
		----------
		dest_folder : str
			Destination folder where scripts should be stored.

		recipe : dict
			Parameters to generate the scripts.

		empty_dest : boolean
			If `True` and if the destination folder already exists, empty it before generating the scripts. If `False` the existence of the folder raises an error.

		Raises
		------
		EmptyListError
			The list of simulations to generate is empty.

		DestinationFolderExistsError
			The destination folder already exists.

		Returns
		-------
		generated_scripts : list
			List of generated scripts, separated: one list per skeleton, in the order they are called.
		'''

		if not(self._simulations_to_generate):
			raise EmptyListError()

		if os.path.isdir(dest_folder):
			if empty_dest:
				for entry in [os.path.join(dest_folder, e) for e in os.listdir(dest_folder)]:
					(shutil.rmtree if os.path.isdir(entry) else os.unlink)(entry)

			else:
				raise DestinationFolderExistsError()

		else:
			os.makedirs(dest_folder)

		if not('max_simulations' in recipe):
			if 'max_subgroups' in recipe:
				recipe['max_simulations'] = ceil(len(self._simulations_to_generate) / recipe['max_subgroups'])

			else:
				recipe['max_simulations'] = len(self._simulations_to_generate)

		else:
			if recipe['max_simulations'] <= 0:
				recipe['max_simulations'] = len(self._simulations_to_generate)

			if 'max_subgroups' in recipe and len(self._simulations_to_generate) / recipe['max_simulations'] > recipe['max_subgroups']:
				recipe['max_simulations'] = ceil(len(self._simulations_to_generate) / recipe['max_subgroups'])

		simulations_sets = [self._simulations_to_generate[k:k+recipe['max_simulations']] for k in range(0, len(self._simulations_to_generate), recipe['max_simulations'])]

		data_lists = {}
		data_variables = {
			'JOBS_OUTPUT_FILENAME': recipe['jobs_output_filename']
		}

		if 'jobs_states_filename' in recipe:
			data_variables['JOBS_STATES_FILENAME'] = recipe['jobs_states_filename']

		if 'data_lists' in recipe:
			data_lists.update(recipe['data_lists'])

		if 'data_variables' in recipe:
			data_variables.update(recipe['data_variables'])

		n_subgroups_skeletons = len(recipe['subgroups_skeletons'])
		n_skeletons = n_subgroups_skeletons + len(recipe['wholegroup_skeletons'])

		skeletons_calls = []
		generated_scripts = [[]] * n_skeletons

		if 'subgroups_skeletons' in recipe:
			skeletons_calls += [
				{
					'skeleton_name_joiner': f'-{k}.',
					'skeletons': enumerate(recipe['subgroups_skeletons']),
					**self.parse(simulations_set)
				}
				for k, simulations_set in enumerate(simulations_sets)
			]

		if 'wholegroup_skeletons' in recipe:
			skeletons_calls.append({
				'skeleton_name_joiner': '.',
				'skeletons': [(n_subgroups_skeletons + j, s) for j, s in enumerate(recipe['wholegroup_skeletons'])],
				**self.parse()
			})

		if not('make_executable' in recipe):
			recipe['make_executable'] = False

		scripts_basedir = dest_folder if not('basedir' in recipe) else recipe['basedir']

		for skeletons_call in skeletons_calls:
			data_lists.update(skeletons_call['data_lists'])
			data_variables.update(skeletons_call['data_variables'])

			if 'data_variables_cases' in recipe:
				for varname, varparams in recipe['data_variables_cases'].items():
					vartest = data_variables[varparams['variable']]
					data_variables[varname] = [value for bound, value in zip(varparams['bounds'], varparams['values']) if bound <= vartest][-1]

			for j, skeleton_name in skeletons_call['skeletons']:
				skeleton_basename_parts = os.path.basename(skeleton_name).rsplit('.skeleton.', maxsplit = 1)
				skeleton_tag = re.sub('[^A-Z_]+', '_', skeleton_basename_parts[0].upper())
				script_name = skeletons_call['skeleton_name_joiner'].join(skeleton_basename_parts)
				script_localpath = os.path.join(dest_folder, script_name)
				script_finalpath = os.path.join(scripts_basedir, script_name)

				self.generateScriptFromSkeleton(skeleton_name, script_localpath, data_lists, data_variables, make_executable = recipe['make_executable'])
				data_variables[skeleton_tag] = script_finalpath

				try:
					data_lists[skeleton_tag].append(script_finalpath)

				except KeyError:
					data_lists[skeleton_tag] = [script_finalpath]

				generated_scripts[j].append({'name': script_name, 'localpath': script_localpath, 'finalpath': script_finalpath})

		return generated_scripts



#!/usr/bin/env python3
# -*- coding: utf-8 -*-

import os
import re
import inspect

from . import jsonfiles
from .errors import *
from . import fixers

class Folder():
	'''
	Base class for each system needing access to the configuration files of a simulations folder.
	Initialize with the simulations folder and load the settings.

	Parameters
	----------
	folder : str
		The simulations folder. Must contain a settings file.

	Raises
	------
	FileNotFoundError
		No `.simulations.conf` file found in folder.
	'''

	def __init__(self, folder):
		self._folder = folder
		self._settings_file = os.path.join(self._folder, '.simulations.conf')

		if not(os.path.isfile(self._settings_file)):
			raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), self._settings_file)

		self._settings = None

		self._fixers_regex_compiled = None
		self._fixers_list = None

	@property
	def folder(self):
		'''
		Return the folder's path.

		Returns
		-------
		path : str
			The path.
		'''

		return self._folder

	@property
	def settings(self):
		'''
		Return the content of the settings file as a dictionary.

		Returns
		-------
		settings : dict
			The folder's settings.
		'''

		if not(self._settings):
			self._settings = jsonfiles.read(self._settings_file)

			if not('fixes' in self._settings):
				self._settings['fixes'] = []

		return self._settings

	@property
	def _fixers_regex(self):
		'''
		Regex to detect whether a function's name corresponds to a value fixer.

		Returns
		-------
		regex : re.Pattern
			The fixers regex.
		'''

		if self._fixers_regex_compiled is None:
			self._fixers_regex_compiled = re.compile(r'^fixer_([A-Za-z0-9_]+)$')

		return self._fixers_regex_compiled

	@property
	def _fixers(self):
		'''
		Get the list of available values fixers.

		Returns
		-------
		fixers : dict
			The values fixers.
		'''

		if self._fixers_list is None:
			self._fixers_list = {}
			self.loadFixersFromModule(fixers)

		return self._fixers_list

	def loadFixersFromModule(self, module):
		'''
		Load all values fixers in a given module.

		Parameters
		----------
		module : Module
			Module (already loaded) where are defined the fixers.
		'''

		for function in inspect.getmembers(module, inspect.isfunction):
			fixer_match = self._fixers_regex.match(function[0])

			if fixer_match:
				self.setFixer(fixer_match.group(1), function[1])

	def setFixer(self, fixer_name, fixer):
		'''
		Set (add or replace) a value fixer.

		Parameters
		----------
		fixer_name : str
			Name of the fixer.

		fixer : function
			Fixer to register.
		'''

		self._fixers[fixer_name] = fixer

	def removeFixer(self, fixer_name):
		'''
		Remove a value fixer.

		Parameters
		----------
		fixer_name : str
			Name of the fixer to remove.
		'''

		if not(fixer_name in self._fixers):
			raise FixerNotFoundError(fixer_name)

		del self._fixers[fixer_name]

	def applyFixers(self, value, before = [], after = []):
		'''
		Fix a value to prevent false duplicates (e.g. this prevents to consider `0.0` and `0` as different values).
		Each item of a list of fixers is either a fixer's name or a list beginning with the fixer's name and followed by the arguments to pass to the fixer.

		Parameters
		----------
		value : mixed
			The value to fix.

		before : list
			List of fixers to apply before the global ones.

		after : list
			List of fixers to apply after the global ones.

		Returns
		-------
		fixed : mixed
			The same value, fixed.

		Raises
		------
		FixerNotFoundError
			The fixer's name has not been found.
		'''

		for fixer in before + self.settings['fixes'] + after:
			if not(type(fixer) is list):
				fixer = [fixer]

			if not(fixer[0] in self._fixers):
				raise FixerNotFoundError(fixer[0])

			value = self._fixers[fixer[0]](value, *fixer[1:])

		return value
